//
//  JemaWrapper.swift
//
//  Copyright © 2022 Anagog Ltd. All rights reserved.
//

import Foundation
import JedAI
import JedAIJEMA
import JedAIPOI
import JedAIReport
import JedAIMetrics
import JedAILambda
import JedAIUI
import CoreMotion
import CoreLocation
import os

public class JemaWrapper {
    
    public static let shared = JemaWrapper()

    /**
     This is the configuratoin of the POI providers that will be used by JEMA.
     Do not touch this unless you specifically get different definitions from Anagog project team.
     - seealso:
     [POI Providers Configuration]
     (https://docs.anagog.com/docs/smart-poi#poi-providers)
     */
    private var poiProvidersConfig: [POIProviderConfig] {
        return [
            POIAlgorithmType4Config(),
            POIAlgorithmType1Config(),
        ]
    }
    
    var jedai = JedAI.shared
    var jedaiJEMA = JedAIJEMA.shared
    var jedaiPoi = JedAIPOI.shared
    var jedaiReport = JedAIReport.shared
    var jedaiMetrics = JedAIMetrics.shared
    var jedaiLambda = JedAILambda.shared
    var jedaiUI = JedAIUI.shared
    
    private var apiKey = ""
    private var userDefinedMetrics: [String: JedAIMetricsModel.Metric.MetricType] = [:]
    private var clientName: String? = nil
    private var postStartup: (() -> ())? = nil
    
    public func configure(apiKey: String, clientName: String? = nil, userDefinedMetrics: [String: JedAIMetricsModel.Metric.MetricType] = [:]) {
        self.apiKey = apiKey
        self.clientName = clientName
        self.userDefinedMetrics = userDefinedMetrics
    }
    
    public func onPostStartup(_ executionBlock: @escaping () -> ()) {
        self.postStartup = executionBlock
    }

    @objc
    public func startup() {
        self.jedai.startup(apiKey: self.apiKey) {
            self.jedaiMetrics.startup(userDefinedMetrics: self.userDefinedMetrics)
            self.jedaiPoi.startup(configs: self.poiProvidersConfig)
            self.jedaiReport.startup()
            self.jedaiLambda.startup()
            self.jedaiJEMA.startup(clientName: self.clientName)
            self.jedaiUI.startup()
            self.postStartup?()
        }
        
        if self.jedaiUI.campaignTriggerActionHandler == nil {
            self.jedaiUI.campaignTriggerActionHandler = { action, userInfo in
                debugPrint("campaignTriggerActionHandler executed with action: \(action ?? ""), userInfo: \(userInfo)")
                SwiftAnagogFlutterPlugin.shared.sendEvent([
                    "action": "onCampaignTriggered",
                    "payload": userInfo
                ])
            }
        }
        
        if self.jedaiUI.campaignNotificationClickActionHandler == nil {
            self.jedaiUI.campaignNotificationClickActionHandler = {action, actionPerformed, userInfo in
                debugPrint("notificationActionExecutionHandler executed with action: \(action ?? ""), success: \(actionPerformed), userInfo: \(userInfo)")
                var payload = userInfo
                payload["executed"] = actionPerformed
                SwiftAnagogFlutterPlugin.shared.sendEvent([
                    "action": "onNotificationClick",
                    "payload": payload
                ])
            }
        }

        if self.jedaiUI.campaignNotificationApprovementHandler == nil {
            self.jedaiUI.campaignNotificationApprovementHandler = { userInfo -> Bool in
                debugPrint("notificationApprovementHandler executed with event: \(userInfo)")
                return true
            }
        }
        
        if self.jedaiReport.microsegmentsSnapshotReportHandler == nil {
            self.jedaiReport.microsegmentsSnapshotReportHandler = { payload in
                debugPrint("microsegmentsSnapshotReportHandler executed with payload: \(payload)")
                SwiftAnagogFlutterPlugin.shared.sendEvent([
                    "action": "onSnapshotReport",
                    "payload": payload
                ])
            }
        }
    }

    /**
        Shutdown JEMA Framework
        Call this method before application termination.
     */
    @objc
    public func shutdown() {
        self.jedaiJEMA.shutdown()
        self.jedaiLambda.shutdown()
        self.jedaiMetrics.shutdown()
        self.jedaiReport.shutdown()
        self.jedaiPoi.shutdown()
        self.jedai.shutdown()
        self.jedaiUI.shutdown()
    }

    /**
        Refresh JEMA Framework
        Call this method during background fetch request or any other trigger, for example application goes to foreground.
     */
    @objc
    public func refresh() {
        self.jedai.refresh()
    }
    
    @objc
    public func didReceiveNotification(_ notification: UNNotification) {
        JedAI.shared.didReceiveNotification(notification)
    }
    
    @objc
    public func didReceiveRemoteNotification(_ userInfo: [AnyHashable: Any]) {
        JedAI.shared.didReceiveRemoteNotification(userInfo)
    }
    
    @objc
    public func updateDeviceToken(_ deviceToken: Data) {
        JedAI.shared.updateDeviceToken(deviceToken)
    }
}

// MARK: - THE FOLLOWING CODE SHOULD BE USED ONLY FOR INTEGRATION PURPOSES

extension JemaWrapper {
   func syncCampaigns() {
       os_log("Sync Campaigns")
       DispatchQueue.global().async {
           JedAICommandEvent(command: "enforceJEMASync").post()
       }
   }
    
   func simulateCampaigns(_ campaignIdentifier: String? = nil) {
       os_log("Simulate Campaigns")
       self.jedaiJEMA.simulateCampaign(campaignIdentifier)
   }
}
