//
//  SwiftAnagogFlutterPlugin.swift
//
//  Copyright © 2022 Anagog Ltd. All rights reserved.
//

import Flutter
import Foundation
import JedAI
import JedAIJEMA
import JedAIMetrics
import CoreMotion
import CoreLocation

public class SwiftAnagogFlutterPlugin: NSObject {
    
    public static let shared = SwiftAnagogFlutterPlugin()
    
    private var channelMethods: FlutterMethodChannel? = nil
    private var channelEvents: FlutterEventChannel? = nil
    private var eventSink: FlutterEventSink? = nil
    private var pendingEventsQueue = DispatchQueue(label: "com.anagog.flutter.pending.events.queue", qos: .utility)
    private var pendingEvents: [[String: Any]] = []
    
    // MARK: - Initialization

    private override init() {
        super.init()
    }
    
    // MARK: - Micromoments
    
    private func fireMicromoment(_ micromoment: Dictionary<String, Any>) {
        guard let identifier = micromoment["identifier"] as? String else {
            return
        }
        var applicationEvent = ApplicationEvent(identifier)
        
        if let textParameters = micromoment["textParameters"] as? Dictionary<String, String> {
            for (name, text) in textParameters {
                applicationEvent = applicationEvent.addParameter(name: name, text: text)
            }
        }
        if let numericParameters = micromoment["numericParameters"] as? Dictionary<String, Float> {
            for (name, number) in numericParameters {
                applicationEvent = applicationEvent.addParameter(name: name, number: number)
            }
        }
        if let locationParameters = micromoment["locationParameters"] as? Dictionary<String, Dictionary<String, CLLocationDegrees>> {
            for (name, rawLocation) in locationParameters {
                guard let latitude = rawLocation["latitude"], let longitude = rawLocation["longitude"] else {
                    continue
                }
                let location = CLLocationCoordinate2D(latitude: latitude, longitude: longitude)
                applicationEvent = applicationEvent.addParameter(name: name, location: location)
            }
        }
        applicationEvent.post()
    }
    
    // MARK: - Microsegments
    
    private func getUserDefinedMetrics(_ callback: FlutterResult) {
        var result = [String: String]()
        for (metricName, metricType) in JedAIMetrics.shared.userDefinedMetrics {
            switch metricType {
            case .string:
                result[metricName] = "String"
            case .decimal:
                result[metricName] = "Decimal"
            case .integer:
                result[metricName] = "Integer"
            case .variant:
                break
            }
        }
        callback(result)
    }
    
    private func setUserDefinedString(_ customName: String, value: String) {
        JedAIMetricsModel.Metric.UserDefinedString.set(customName: customName, value: value)
    }
    
    private func setUserDefinedInteger(_ customName: String, value: Int) {
        JedAIMetricsModel.Metric.UserDefinedInteger.set(customName: customName, value: value)
    }
    
    private func setUserDefinedDecimal(_ customName: String, value: Double) {
        JedAIMetricsModel.Metric.UserDefinedDecimal.set(customName: customName, value: value)
    }
    
    private func getUserDefinedString(_ customName: String, callback: FlutterResult) {
        let value = JedAIMetricsModel.Metric.UserDefinedString.get(customName: customName) as String
        callback(value)
    }
    
    private func getUserDefinedInteger(_ customName: String, callback: FlutterResult) {
        let value = JedAIMetricsModel.Metric.UserDefinedInteger.get(customName: customName) as Int
        callback(value)
    }
    
    private func getUserDefinedDecimal(_ customName: String, callback: FlutterResult) {
        let value = JedAIMetricsModel.Metric.UserDefinedDecimal.get(customName: customName) as Double
        callback(value)
    }
    
    // MARK: - Campaigns
    
    private func downloadCampaigns() {
        JemaWrapper.shared.syncCampaigns()
    }
    
    private func simulateCampaign(_ campaignIdentifier: String? = nil) {
        JemaWrapper.shared.simulateCampaigns(campaignIdentifier)
    }
    
    // MARK: - Sharing
    
    private func share() {
        DispatchQueue.main.async {
            ShareService.shareDatabaseAndLogs()
        }
    }
    
}


// MARK: - FlutterPlugin protocol

extension SwiftAnagogFlutterPlugin: FlutterPlugin {
    
    private static let methodsChannelName = "com.anagog.flutter/method_channel"
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let binaryMessenger = registrar.messenger()
        
        if shared.channelMethods == nil {
            shared.channelMethods = FlutterMethodChannel(name: methodsChannelName, binaryMessenger: binaryMessenger)
            shared.channelMethods?.setMethodCallHandler(shared.handle(_:result:))
        }
        
        if shared.channelEvents == nil {
            shared.channelEvents = FlutterEventChannel(name: eventsChannelName, binaryMessenger: binaryMessenger)
            shared.channelEvents?.setStreamHandler(shared)
        }
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "syncCampaigns":
            self.downloadCampaigns()
            result(nil)
        case "simulateCampaign":
            self.simulateCampaign(call.arguments as? String)
            result(nil)
        case "getUserDefinedString":
            guard let key = call.arguments as? String else {
                result(nil)
                return
            }
            self.getUserDefinedString(key, callback: result)
        case "getUserDefinedDecimal":
            guard let key = call.arguments as? String else {
                result(nil)
                return
            }
            self.getUserDefinedDecimal(key, callback: result)
        case "getUserDefinedInteger":
            guard let key = call.arguments as? String else {
                result(nil)
                return
            }
            self.getUserDefinedInteger(key, callback: result)
        case "setUserDefinedString":
            guard let arguments = call.arguments as? [String: Any],
                  let key = arguments["key"] as? String,
                  let value = arguments["value"] as? String
            else {
                result(nil)
                return
            }
            self.setUserDefinedString(key, value: value)
            result(nil)
        case "setUserDefinedDecimal":
            guard let arguments = call.arguments as? [String: Any],
                  let key = arguments["key"] as? String,
                  let value = arguments["value"] as? Double
            else {
                result(nil)
                return
            }
            self.setUserDefinedDecimal(key, value: value)
            result(nil)
        case "setUserDefinedInteger":
            guard let arguments = call.arguments as? [String: Any],
                  let key = arguments["key"] as? String,
                  let value = arguments["value"] as? Int
            else {
                result(nil)
                return
            }
            
            self.setUserDefinedInteger(key, value: value)
            result(nil)
        case "getVersion":
            result(JedAI.shared.api.version)
        case "getUserDefinedStat":
            self.getUserDefinedMetrics(result)
        case "fireMicromoments":
            guard let micromoment = call.arguments as? [String: Any] else {
                result(FlutterError(code: "", message: "Argument format is wrong", details: nil))
                return
            }
            self.fireMicromoment(micromoment)
            result(nil)
        case "share":
            self.share()
            result(nil)
        default:
            result(FlutterMethodNotImplemented)
        }
    }
    
}


// MARK: - FlutterStreamHandler protocol

extension SwiftAnagogFlutterPlugin: FlutterStreamHandler {
    
    private static let eventsChannelName = methodsChannelName + "/events"
    
    internal func sendEvent(_ notificationInfo: [String: Any]) {
        if let eventSink = self.eventSink {
            os_log_file("JEMAEmitter send event")
            eventSink(notificationInfo)
        }
        else {
            os_log_file("JEMAEmitter not activated yet. Event was added to the queue")
            pendingEventsQueue.sync {
                pendingEvents.append(notificationInfo)
            }
        }
    }
    
    public func onListen(withArguments arguments: Any?, eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        os_log_file("JEMAEmitter start listen")
        self.eventSink = events
        self.pendingEventsQueue.sync {
            os_log_file("JEMAEmitter there are %d pending event(s)", self.pendingEvents.count)
            while self.pendingEvents.count > 0 {
                let event = self.pendingEvents.removeFirst()
                self.sendEvent(event)
            }
        }
        return nil
    }
    
    public func onCancel(withArguments arguments: Any?) -> FlutterError? {
        os_log_file("JEMAEmitter stop listen")
        return nil
    }
    
}
